@@ -21,8 +21,13 @@
 #ifndef YASH_COMMON_H
 #define YASH_COMMON_H
 
-#if HAVE_CONFIG_H
-# include "config.h"
+#include "config.h"
+
+#ifndef _POSIX_C_SOURCE
+# define _POSIX_C_SOURCE 200112L
+#endif
+#ifndef _XOPEN_SOURCE
+# define _XOPEN_SOURCE 600
 #endif
 
 #ifndef __GNUC__
