@@ -426,15 +426,14 @@ FILE *open_histfile(void)
 	return NULL;
     }
 
-    int newfd = copy_as_shellfd(fd);
-    xclose(fd);
-    if (newfd < 0)
+    fd = move_to_shellfd(fd);
+    if (fd < 0)
 	return NULL;
 
-    FILE *f = fdopen(newfd, "r+");
+    FILE *f = fdopen(fd, "r+");
     if (!f) {
-	remove_shellfd(newfd);
-	xclose(newfd);
+	remove_shellfd(fd);
+	xclose(fd);
     }
     return f;
 }
@@ -911,7 +910,8 @@ refresh:
     }
 }
 
-/* If the history file is open, close it. */
+/* Closes the history file after writing info on the shell process that is
+ * exiting. */
 void finalize_history(void)
 {
     if (!is_interactive_now || !histfile)
@@ -932,6 +932,16 @@ void finalize_history(void)
     }
 }
 
+/* Closes the history file if open. */
+void close_history_file(void)
+{
+    hist_lock = false;
+    if (histfile == NULL)
+	return;
+    fclose(histfile);
+    histfile = NULL;
+}
+
 /* Adds the specified `line' to the history.
  * Must not be called while the history is locked.
  * If `line' contains newlines, `line' is separated into multiple entries.
@@ -1003,10 +1013,9 @@ void really_add_history(const wchar_t *line)
  * history file should be flushed after this function. */
 void remove_dups(const char *line)
 {
-    histentry_T *e = Histlist;
-    histentry_T *saveprev = e->Prev;
+    histentry_T *saveprev = histlist.Newest;
     for (unsigned i = histrmdup; i > 0; i--) {
-	e = saveprev;
+	histentry_T *e = saveprev;
 	saveprev = e->Prev;
 	if (e == Histlist)
 	    break;
@@ -1448,26 +1457,27 @@ error1:
 		    editor ? editor : L"${FCEDIT:-ed}", temp);
 	}
 
+	int fd;
 	if (laststatus != Exit_SUCCESS) {
 	    xerror(0, Ngt("editor returned non-zero status"));
-	    f = NULL;
+	    fd = -1;
 	} else {
-	    f = reopen_with_shellfd(fopen(temp, "r"), "r", false);
-	    if (f == NULL)
+	    fd = move_to_shellfd(open(temp, O_RDONLY));
+	    if (fd < 0)
 		xerror(errno, Ngt("cannot read command from `%s'"), temp);
 	}
 	if (unlink(temp) < 0)
 	    xerror(errno, Ngt("cannot remove temporary file `%s'"), temp);
 	free(temp);
 
-	if (f == NULL)
+	if (fd < 0)
 	    return Exit_FAILURE;
 
-	int fd = fileno(f);
+	f = fdopen(fd, "r");
 	fc_read_history(f, quiet);
-	rewind(f);
+	lseek(fd, 0, SEEK_SET);
 	laststatus = savelaststatus;
-	exec_input(f, "fc", false, true, false);
+	exec_input(fd, "fc", false, true, false);
 	remove_shellfd(fd);
 	fclose(f);
 	return laststatus;
