@@ -185,6 +185,7 @@ void le_complete(le_compresult_T lecr)
 	le_state = LE_STATE_SUSPENDED_COMPDEBUG;
 	le_compdebug("completion start");
     } else {
+	le_state = LE_STATE_ACTIVE_COMPLETING;
 	le_allow_terminal_signal(true);
     }
 
@@ -205,14 +206,18 @@ void le_complete(le_compresult_T lecr)
 
     if (le_state == LE_STATE_SUSPENDED_COMPDEBUG) {
 	le_compdebug("completion end");
-	le_setupterm(false);
+	le_setupterm(true);
 	le_set_terminal();
-	le_state = LE_STATE_ACTIVE;
     } else {
+	assert(le_state == LE_STATE_ACTIVE_COMPLETING);
 	le_allow_terminal_signal(false);
-	if (is_interrupted())
-	    le_display_clear(false);  /* redraw if interrupted */
+
+	/* the terminal size may have been changed during completion, so we
+	 * re-check the terminal state here. */
+	le_display_clear(false);
+	le_setupterm(true);
     }
+    le_state = LE_STATE_ACTIVE;
 }
 
 /* An `le_compresult_T' function that does nothing. */
