@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* option.h: option settings */
-/* (C) 2007-2009 magicant */
+/* (C) 2007-2010 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -49,7 +49,8 @@ extern _Bool shopt_braceexpand;
 extern _Bool shopt_noclobber;
 extern enum shopt_lineedit_T shopt_lineedit;
 extern enum shopt_yesnoauto_T shopt_le_convmeta;
-extern _Bool shopt_le_visiblebell, shopt_le_promptsp, shopt_le_alwaysrp;
+extern _Bool shopt_le_visiblebell, shopt_le_promptsp, shopt_le_alwaysrp,
+       shopt_le_compdebug;
 
 #define SHELLSET_OPTIONS L"abefhmnuvxC"
 
