@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* redir.c: manages file descriptors and provides functions for redirections */
-/* (C) 2007-2009 magicant */
+/* (C) 2007-2010 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -78,7 +78,8 @@ int xclose(int fd)
  * `xclose' is called before `dup2'. */
 int xdup2(int oldfd, int newfd)
 {
-    xclose(newfd);
+    if (oldfd != newfd)
+	xclose(newfd);
     while (dup2(oldfd, newfd) < 0) {
 	switch (errno) {
 	case EINTR:
@@ -218,53 +219,39 @@ void clear_shellfds(bool leavefds)
  * On error, `errno' is set and -1 is returned. */
 int copy_as_shellfd(int fd)
 {
+    int newfd;
+
 #ifdef F_DUPFD_CLOEXEC
-    int newfd = fcntl(fd, F_DUPFD_CLOEXEC, shellfdmin);
-#else
-    int newfd = fcntl(fd, F_DUPFD, shellfdmin);
+    /* Even if the F_DUPFD_CLOEXEC flag is defined in the <fcntl.h> header, the
+     * OS kernel may not support it. We fall back on the normal F_DUPFD-F_SETFD
+     * sequence if the F_DUPFD_CLOEXEC flag is rejected. */
+    static bool dupfd_cloexec_ok = true;
+    if (dupfd_cloexec_ok) {
+	newfd = fcntl(fd, F_DUPFD_CLOEXEC, shellfdmin);
+	if (newfd >= 0 || errno != EINVAL)
+	    goto finish;
+	dupfd_cloexec_ok = false;
+    }
 #endif
-    if (newfd >= 0) {
-#ifndef F_DUPFD_CLOEXEC
+
+    newfd = fcntl(fd, F_DUPFD, shellfdmin);
+    if (newfd >= 0)
 	fcntl(newfd, F_SETFD, FD_CLOEXEC);
+#ifdef F_DUPFD_CLOEXEC
+finish:
 #endif
+    if (newfd >= 0)
 	add_shellfd(newfd);
-    }
     return newfd;
 }
 
-/* Duplicates the underlining file descriptor of the specified stream.
- * The original stream is closed whether successful or not.
- * A new stream is open with the new FD using `fdopen' and returned.
- * The new stream's underlining file descriptor is registered as a shell FD and,
- * if `nonblock' is true, set to non-blocking.
- * If NULL is given, this function just returns NULL without doing anything. */
-FILE *reopen_with_shellfd(FILE *f, const char *mode, bool nonblock)
-{
-    if (!f)
-	return NULL;
-
-    int newfd = copy_as_shellfd(fileno(f));
-    fclose(f);
-    if (newfd < 0)
-	return NULL;
-    if (nonblock && !set_nonblocking(newfd)) {
-	xclose(newfd);
-	return NULL;
-    }
-    return fdopen(newfd, mode);
-}
-
 /* Opens `ttyfd'.
  * On failure, an error message is printed and `do_job_control' is set to false.
  */
 void open_ttyfd(void)
 {
     if (ttyfd < 0) {
-	int fd = open("/dev/tty", O_RDWR);
-	if (fd >= 0) {
-	    ttyfd = copy_as_shellfd(fd);
-	    xclose(fd);
-	}
+	ttyfd = move_to_shellfd(open("/dev/tty", O_RDWR));
 	if (ttyfd < 0) {
 	    xerror(errno, Ngt("cannot open `%s'"), "/dev/tty");
 	    xerror(0, Ngt("job control disabled"));
@@ -889,7 +876,7 @@ void maybe_redirect_stdin_to_devnull(void)
 
     xclose(STDIN_FILENO);
     fd = open("/dev/null", O_RDONLY);
-    if (fd > 0) {
+    if (fd >= 0 && fd != STDIN_FILENO) {  // shouldn't happen, but just in case
 	xdup2(fd, STDIN_FILENO);
 	xclose(fd);
     }
