@@ -35,6 +35,7 @@
 # include <libintl.h>
 #endif
 #include "builtin.h"
+#include "configm.h"
 #include "exec.h"
 #include "expand.h"
 #include "hashtable.h"
@@ -47,9 +48,10 @@
 #include "strbuf.h"
 #include "util.h"
 #include "variable.h"
-#include "version.h"
+#include "xfnmatch.h"
 #include "yash.h"
 #if YASH_ENABLE_LINEEDIT
+# include "lineedit/complete.h"
 # include "lineedit/lineedit.h"
 # include "lineedit/terminfo.h"
 #endif
@@ -58,7 +60,9 @@
 static const wchar_t *const path_variables[PA_count] = {
     [PA_PATH]     = L VAR_PATH,
     [PA_CDPATH]   = L VAR_CDPATH,
-    [PA_MAILPATH] = L VAR_MAILPATH,
+#if YASH_ENABLE_LINEEDIT
+    [PA_COMPPATH] = L VAR_YASH_COMPPATH,
+#endif
 };
 
 
@@ -80,8 +84,8 @@ typedef struct environ_T {
 /* An environment whose `is_temporary' is true is used for temporary variables. 
  * Temporary variables may be exported but cannot be readonly.
  * The elements of `paths' are arrays of the pathnames contained in the
- * $PATH, $CDPATH and $MAILPATH variables. They are NULL if the corresponding
- * variables are not set. */
+ * $PATH, $CDPATH and $YASH_COMPPATH variables. They are NULL if the
+ * corresponding variables are not set. */
 #define VAR_positional "="
 
 /* flags for variable attributes */
@@ -165,6 +169,10 @@ static unsigned next_random(void);
 static void variable_set(const wchar_t *name, variable_T *var)
     __attribute__((nonnull(1)));
 
+static char **convert_path_array(void **ary)
+    __attribute__((malloc,warn_unused_result));
+static void add_to_list_no_dup(plist_T *list, char *s)
+    __attribute__((nonnull(1)));
 static void reset_path(path_T name, variable_T *var);
 
 static void funcfree(function_T *f);
@@ -237,8 +245,8 @@ void init_variables(void)
     current_env->parent = NULL;
     current_env->is_temporary = false;
     ht_init(&current_env->contents, hashwcs, htwcscmp);
-    for (size_t i = 0; i < PA_count; i++)
-	current_env->paths[i] = NULL;
+//    for (size_t i = 0; i < PA_count; i++)
+//	current_env->paths[i] = NULL;
 
     ht_init(&functions, hashwcs, htwcscmp);
 
@@ -260,6 +268,10 @@ void init_variables(void)
 	varkvfree(ht_set(&current_env->contents, we, v));
     }
 
+    /* initialize path according to $PATH etc. */
+    for (size_t i = 0; i < PA_count; i++)
+	current_env->paths[i] = decompose_paths(getvar(path_variables[i]));
+
     /* set $IFS */
     set_variable(L VAR_IFS, xwcsdup(DEFAULT_IFS), SCOPE_GLOBAL, false);
 
@@ -319,13 +331,13 @@ void init_variables(void)
 	random_active = false;
     }
 
+    /* set $YASH_COMPPATH */
+    set_variable(L VAR_YASH_COMPPATH, xwcsdup(L YASH_DATADIR "/completion"),
+	    SCOPE_GLOBAL, false);
+
     /* set $YASH_VERSION */
     set_variable(L VAR_YASH_VERSION, xwcsdup(L PACKAGE_VERSION),
 	    SCOPE_GLOBAL, false);
-
-    /* initialize path according to $PATH/CDPATH */
-    for (size_t i = 0; i < PA_count; i++)
-	current_env->paths[i] = decompose_paths(getvar(path_variables[i]));
 }
 
 /* Reset the value of $PWD if
@@ -1006,10 +1018,6 @@ void variable_set(const wchar_t *name, variable_T *var)
 	    le_need_term_update = true;
 #endif
 	break;
-    case L'M':
-	if (wcscmp(name, L VAR_MAILPATH) == 0)
-	    reset_path(PA_MAILPATH, var);
-	break;
     case L'P':
 	if (wcscmp(name, L VAR_PATH) == 0) {
 	    clear_cmdhash();
@@ -1034,6 +1042,10 @@ void variable_set(const wchar_t *name, variable_T *var)
 	if (wcscmp(name, L VAR_TERM) == 0)
 	    le_need_term_update = true;
 	break;
+    case L'Y':
+	if (wcscmp(name, L VAR_YASH_COMPPATH) == 0)
+	    reset_path(PA_COMPPATH, var);
+	break;
 #endif /* YASH_ENABLE_LINEEDIT */
     }
 }
@@ -1050,38 +1062,53 @@ char **decompose_paths(const wchar_t *paths)
     if (!paths)
 	return NULL;
 
-    wchar_t wpath[wcslen(paths) + 1];
-    wcscpy(wpath, paths);
+    plist_T list;
+    pl_init(&list);
+
+    const wchar_t *colon;
+    while ((colon = wcschr(paths, L':')) != NULL) {
+	add_to_list_no_dup(&list, malloc_wcsntombs(paths, colon - paths));
+	paths = colon + 1;
+    }
+    add_to_list_no_dup(&list, malloc_wcstombs(paths));
+
+    return (char **) pl_toary(&list);
+}
+
+/* Converts an array of wide strings into an newly-malloced array of multibyte
+ * strings.
+ * If `paths' is NULL, NULL is returned. */
+char **convert_path_array(void **ary)
+{
+    if (!ary)
+	return NULL;
 
     plist_T list;
     pl_init(&list);
 
-    /* add each element to `list', replacing each L':' with L'\0' in `wpath'. */
-    pl_add(&list, wpath);
-    for (wchar_t *w = wpath; *w; w++) {
-	if (*w == L':') {
-	    *w = L'\0';
-	    pl_add(&list, w + 1);
-	}
-    }
-
-    /* remove duplicates */
-    for (size_t i = 0; i < list.length; i++)
-	for (size_t j = list.length; --j > i; )
-	    if (wcscmp(list.contents[i], list.contents[j]) == 0)
-		pl_remove(&list, j, 1);
-
-    /* convert each element back to multibyte string */
-    for (size_t i = 0; i < list.length; i++) {
-	list.contents[i] = malloc_wcstombs(list.contents[i]);
-	/* We actually assert this conversion always succeeds, but... */
-	if (!list.contents[i])
-	    list.contents[i] = xstrdup("");
+    while (*ary) {
+	add_to_list_no_dup(&list, malloc_wcstombs(*ary));
+	ary++;
     }
 
     return (char **) pl_toary(&list);
 }
 
+/* If `s' is non-NULL and not contained in `list', add `s' to list.
+ * Otherwise, frees `s'. */
+void add_to_list_no_dup(plist_T *list, char *s)
+{
+    if (s) {
+	for (size_t i = 0; i < list->length; i++) {
+	    if (strcmp(s, list->contents[i]) == 0) {
+		free(s);
+		return;
+	    }
+	}
+	pl_add(list, s);
+    }
+}
+
 /* Reconstructs the path array of the specified variable in the environment.
  * `var' may be NULL. */
 void reset_path(path_T name, variable_T *var)
@@ -1091,20 +1118,12 @@ void reset_path(path_T name, variable_T *var)
 
 	variable_T *v = ht_get(&env->contents, path_variables[name]).value;
 	if (v) {
-	    plist_T list;
 	    switch (v->v_type & VF_MASK) {
 		case VF_SCALAR:
 		    env->paths[name] = decompose_paths(v->v_value);
 		    break;
 		case VF_ARRAY:
-		    pl_initwithmax(&list, v->v_valc);
-		    for (size_t i = 0; i < v->v_valc; i++) {
-			char *p = malloc_wcstombs(v->v_vals[i]);
-			if (!p)
-			    p = xstrdup("");
-			pl_add(&list, p);
-		    }
-		    env->paths[name] = (char **) pl_toary(&list);
+		    env->paths[name] = convert_path_array(v->v_vals);
 		    break;
 	    }
 	    if (v == var)
@@ -1252,6 +1271,56 @@ void tryhash_word_as_command(const wordunit_T *w)
     }
 }
 
+#if YASH_ENABLE_LINEEDIT
+
+/* Generates completion candidates for variable names that match the glob
+ * pattern in the specified context. */
+/* The prototype of this function is declared in "lineedit/complete.h". */
+void generate_variable_candidates(le_candgentype_T type, le_context_T *context)
+{
+    if (!(type & CGT_VARIABLE))
+	return;
+
+    le_compdebug("adding variables for pattern \"%ls\"", context->pattern);
+    if (!le_compile_cpattern(context))
+	return;
+
+    size_t i = 0;
+    kvpair_T kv;
+    while ((kv = ht_next(&first_env->contents, &i)).key != NULL) {
+	const wchar_t *name = kv.key;
+	const variable_T *var = kv.value;
+	switch (var->v_type & VF_MASK) {
+	    case VF_SCALAR:  if (!(type & CGT_SCALAR)) continue;  break;
+	    case VF_ARRAY:   if (!(type & CGT_ARRAY))  continue;  break;
+	}
+	if (name[0] != L'='
+		&& xfnm_wmatch(context->cpattern, name).start != (size_t) -1)
+	    le_new_candidate(CT_VAR, xwcsdup(name), NULL);
+    }
+}
+
+/* Generates completion candidates for function names that match the glob
+ * pattern in the specified context. */
+/* The prototype of this function is declared in "lineedit/complete.h". */
+void generate_function_candidates(le_candgentype_T type, le_context_T *context)
+{
+    if (!(type & CGT_FUNCTION))
+	return;
+
+    le_compdebug("adding functions for pattern \"%ls\"", context->pattern);
+    if (!le_compile_cpattern(context))
+	return;
+
+    size_t i = 0;
+    const wchar_t *name;
+    while ((name = ht_next(&functions, &i).key) != NULL)
+	if (xfnm_wmatch(context->cpattern, name).start != (size_t) -1)
+	    le_new_command_candidate(xwcsdup(name));
+}
+
+#endif /* YASH_ENABLE_LINEEDIT */
+
 
 /********** Directory Stack **********/
 
@@ -1374,7 +1443,7 @@ static bool set_optind(unsigned long optind, unsigned long optsubind);
 static inline bool set_optarg(const wchar_t *value);
 static bool set_to(const wchar_t *varname, wchar_t value)
     __attribute__((nonnull));
-static bool read_input(xwcsbuf_T *buf, bool noescape)
+static bool read_with_prompt(xwcsbuf_T *buf, bool noescape)
     __attribute__((nonnull));
 static void split_and_assign_array(const wchar_t *name, wchar_t *values,
 	const wchar_t *ifs, bool raw)
@@ -2455,7 +2524,7 @@ int read_builtin(int argc, void **argv)
 
     /* read input and remove trailing newline */
     wb_init(&buf);
-    if (!read_input(&buf, raw)) {
+    if (!read_with_prompt(&buf, raw)) {
 	wb_destroy(&buf);
 	return Exit_FAILURE;
     }
@@ -2504,78 +2573,76 @@ print_usage:
 /* Reads input from the standard input and, if `noescape' is false, remove line
  * continuations.
  * The trailing newline and all other backslashes are not removed. */
-bool read_input(xwcsbuf_T *buf, bool noescape)
+bool read_with_prompt(xwcsbuf_T *buf, bool noescape)
 {
-    if (noescape)
-	return read_line_from_stdin(buf, false);
-
-    bool cont;
     bool first = true;
-    do {
-	size_t index = buf->length;
+    size_t index;
 
-	if (is_interactive_now && isatty(STDIN_FILENO)) {
-	    /* if the shell is interactive and the input is from the terminal,
-	     * then print a prompt and try to use line-editing. */
+read_input:
+    index = buf->length;
+    if (is_interactive_now && isatty(STDIN_FILENO)) {
+	/* if the shell is interactive and the input is from the terminal,
+	 * then print a prompt and try to use line-editing. */
 
-	    struct promptset_T prompt;
-	    if (first) {
-		prompt.main   = xwcsdup(L"");
-		prompt.right  = xwcsdup(L"");
-		prompt.styler = xwcsdup(L"");
-	    } else {
-		prompt = get_prompt(2);
-	    }
+	struct promptset_T prompt;
+	if (first) {
+	    prompt.main   = xwcsdup(L"");
+	    prompt.right  = xwcsdup(L"");
+	    prompt.styler = xwcsdup(L"");
+	} else {
+	    prompt = get_prompt(2);
+	}
 
 #if YASH_ENABLE_LINEEDIT
-	    if (shopt_lineedit != shopt_nolineedit) {
-		wchar_t *line;
-		inputresult_T result = le_readline(prompt, &line);
+	if (shopt_lineedit != shopt_nolineedit) {
+	    wchar_t *line;
+	    inputresult_T result = le_readline(prompt, &line);
 
-		if (result != INPUT_ERROR) {
-		    free_prompt(prompt);
-		    switch (result) {
-			case INPUT_OK:
-			    /* ignore lines after the first one */
-			    wcschr(line, L'\n')[1] = L'\0';
-			    wb_catfree(buf, line);
-			    /* falls thru! */
-			case INPUT_EOF:
-			    goto read;
-			case INPUT_INTERRUPTED:
-			    set_interrupted();
-			    return false;
-			case INPUT_ERROR:
-			    assert(false);
-		    }
+	    if (result != INPUT_ERROR) {
+		free_prompt(prompt);
+		switch (result) {
+		    case INPUT_OK:
+			wb_catfree(buf, line);
+			/* falls thru! */
+		    case INPUT_EOF:
+			goto done;
+		    case INPUT_INTERRUPTED:
+			set_interrupted();
+			return false;
+		    case INPUT_ERROR:
+			assert(false);
 		}
 	    }
+	}
 #endif /* YASH_ENABLE_LINEEDIT */
 
-	    bool result2;
-	    print_prompt(prompt.main);
-	    print_prompt(prompt.styler);
-	    result2 = read_line_from_stdin(buf, false);
-	    print_prompt(PROMPT_RESET);
-	    free_prompt(prompt);
-	    if (!result2)
-		return false;
-	} else {
-	    if (!read_line_from_stdin(buf, false))
-		return false;
-	}
+	inputresult_T result2;
+	print_prompt(prompt.main);
+	print_prompt(prompt.styler);
+	result2 = read_input(buf, stdin_input_file_info, false);
+	print_prompt(PROMPT_RESET);
+	free_prompt(prompt);
+	if (result2 == INPUT_ERROR)
+	    return false;
+    } else {
+	if (read_input(buf, stdin_input_file_info, false) == INPUT_ERROR)
+	    return false;
+    }
+
 #if YASH_ENABLE_LINEEDIT
-read:
+done:
 #endif
-	first = false;
-	cont = false;
+    first = false;
+    if (!noescape) {
+	/* treat escapes */
 	while (index < buf->length) {
 	    if (buf->contents[index] == L'\\') {
 		if (buf->contents[index + 1] == L'\n') {
 		    wb_remove(buf, index, 2);
-		    cont = true;
-		    assert(index == buf->length);
-		    break;
+		    if (index >= buf->length)
+			goto read_input;
+		    else
+			continue;
 		} else {
 		    index += 2;
 		    continue;
@@ -2583,7 +2650,8 @@ read:
 	    }
 	    index++;
 	}
-    } while (cont);
+    }
+
     return true;
 }
 
