@@ -65,12 +65,16 @@ extern char **environ;
 #define VAR_RANDOM                    "RANDOM"
 #define VAR_TERM                      "TERM"
 #define VAR_YASH_AFTER_CD             "YASH_AFTER_CD"
+#define VAR_YASH_COMPPATH             "YASH_COMPPATH"
 #define VAR_YASH_LE_TIMEOUT           "YASH_LE_TIMEOUT"
 #define VAR_YASH_VERSION              "YASH_VERSION"
 #define L                             L""
 
 typedef enum path_T {
-    PA_PATH, PA_CDPATH, PA_MAILPATH,
+    PA_PATH, PA_CDPATH,
+#if YASH_ENABLE_LINEEDIT
+    PA_COMPPATH,
+#endif
     PA_count,
 } path_T;
 
